#ifndef IDENTUSER_H
#define IDENTUSER_H

#ifndef __
# ifdef __STDC__
#  define __(args) args
# else
#  define __(args) ()
# endif
#endif

extern unsigned short ident_tcpport;
extern char *ident_xline();
extern int ident_fd __((int fd, struct in_addr *in, unsigned short *local, unsigned short *remote));
extern int ident_fd2 __((int fd, struct in_addr *inlocal, struct in_addr *inremote, unsigned short *local, unsigned short *remote));
extern int ident_tcpsock __((struct in_addr *inlocal, struct in_addr *inremote));
/* extern char *ident_tcpuser(); */
extern char *ident_tcpuser2 __((/*struct in_addr *inlocal, struct in_addr *inremote, unsigned short local, unsigned short remote*/));
extern char *ident_tcpuser3 __((/*struct in_addr *inlocal, struct in_addr *inremote, unsigned short local, unsigned short remote, int timeout*/));
extern char *ident_sockuser __((/*int s, unsigned short local, unsigned short remote*/));

#endif
