extern void type();

void
debug_report(verb,remotehost,buf)
int verb;
char *remotehost, *buf;
{
#if 0	/* From  UNIDO, though a bit vulgar.. */
  if (!verb) {
    type(423,"To See Figure 1, go into VERBose mode first.");
    return;
  }
  type(-521,"    ---------------------------------");
  type(-521,"    !               _               !");
  type(-521,"    !              { }              !");
  type(-521,"    !              | |              !");
  type(-521,"    !              | |              !");
  type(-521,"    !           .-.! !.-.           !");
  type(-521,"    !         .-!  ! !  !.-.        !");
  type(-521,"    !         ! !       !  ;        !");
  type(-521,"    !         \\           ;         !");
  type(-521,"    !          \\         ;          !");
  type(-521,"    !           !       :           !");
  type(-521,"    !           !       :           !");
  type(-521,"    !           !       :           !");
  type(-521,"    !                               !");
  type(-521,"    !_______________________________!");
  type(-521,"                 Figure 1.           ");
  type(-521,"");
  type(-521,"");
  type(521,"See Figure 1, %s.",remotehost);
  exit(0);
#endif
  if (!verb) {
    type(423,"Must be VERBose to use DEBUG");
    return;
  }
  if (strlen(buf)<8) {
    type(423,"Umm..  DEBUG <password>");
    return;
  }
  type(-521,"Ex cuse me ?  You try to exploit the DEBUG-hole on this hosts sendmail ?");
  type( 521,"Unfortunately (for you) we do not have sendmail...");
#if 0
  type(-521,"Do following, and know what we think of you:");
  type(-521,"   telnet mail.Germany.EU.net smtp");
  type(-521,"   220 ...");
  type(-521,"   VERB");
  type(-521,"   200 ...");
  type( 521,"   DEBUG");
#endif
  exit(0);
}
